@@ -3,7 +3,7 @@ import sys
 import re
 
 
-class UsageMessageError(Exception):
+class DocoptLanguageError(Exception):
 
     """Error in construction of usage-message by developer."""
 
@@ -151,6 +151,7 @@ class Option(Pattern):
         assert argcount in (0, 1)
         self.short, self.long = short, long
         self.argcount, self.value = argcount, value
+        self.value = None if value == False and argcount else value  # HACK
 
     @classmethod
     def parse(class_, option_description):
@@ -166,7 +167,7 @@ class Option(Pattern):
                 argcount = 1
         if argcount:
             matched = re.findall('\[default: (.*)\]', description, flags=re.I)
-            value = matched[0] if matched else False
+            value = matched[0] if matched else None
         return class_(short, long, argcount, value)
 
     def match(self, left, collected=None):
@@ -255,20 +256,17 @@ class Either(Pattern):
         return False, left, collected
 
 
-class TokenStream(object):
+class TokenStream(list):
 
     def __init__(self, source, error):
-        self.s = source.split() if type(source) is str else source
+        self += source.split() if type(source) is str else source
         self.error = error
 
-    def __iter__(self):
-        return iter(self.s)
+    def move(self):
+        return self.pop(0) if len(self) else None
 
-    def move(self, default=None):
-        return self.s.pop(0) if len(self.s) else default
-
-    def current(self, default=None):
-        return self.s[0] if len(self.s) else default
+    def current(self):
+        return self[0] if len(self) else None
 
 
 def parse_long(tokens, options):
@@ -276,7 +274,12 @@ def parse_long(tokens, options):
     value = None if eq == value == '' else value
     opt = [o for o in options if o.long and o.long.startswith(raw)]
     if len(opt) < 1:
-        raise tokens.error('%s is not recognized' % raw)
+        if tokens.error is DocoptExit:
+            raise tokens.error('%s is not recognized' % raw)
+        else:
+            o = Option(None, raw, (1 if eq == '=' else 0))
+            options.append(o)
+            return [o]
     if len(opt) > 1:
         raise tokens.error('%s is not a unique prefix: %s?' %
                          (raw, ', '.join('%s' % o.long for o in opt)))
@@ -302,7 +305,14 @@ def parse_shorts(tokens, options):
             raise tokens.error('-%s is specified ambiguously %d times' %
                               (raw[0], len(opt)))
         if len(opt) < 1:
-            raise tokens.error('-%s is not recognized' % raw[0])
+            if tokens.error is DocoptExit:
+                raise tokens.error('-%s is not recognized' % raw[0])
+            else:
+                o = Option('-' + raw[0], None)
+                options.append(o)
+                parsed.append(o)
+                raw = raw[1:]
+                continue
         opt = copy(opt[0])
         raw = raw[1:]
         if opt.argcount == 0:
@@ -314,13 +324,13 @@ def parse_shorts(tokens, options):
                 raw = tokens.move()
             value, raw = raw, ''
         opt.value = value
-        parsed += [opt]
+        parsed.append(opt)
     return parsed
 
 
 def parse_pattern(source, options):
     tokens = TokenStream(re.sub(r'([\[\]\(\)\|]|\.\.\.)', r' \1 ', source),
-                         UsageMessageError)
+                         DocoptLanguageError)
     result = parse_expr(tokens, options)
     if tokens.current() is not None:
         raise tokens.error('unexpected ending: %r' % ' '.join(tokens))
@@ -377,12 +387,9 @@ def parse_atom(tokens, options):
         if tokens.move() != ']':
             raise tokens.error("Unmatched '['")
         return result
-    elif token == '--':
-        tokens.move()
-        return []  # allow "usage: prog [-o] [--] <arg>"
-    elif token.startswith('--'):
+    elif token.startswith('--') and token != '--':
         return parse_long(tokens, options)
-    elif token.startswith('-'):
+    elif token.startswith('-') and token not in ('-', '--'):
         return parse_shorts(tokens, options)
     elif token.startswith('<') and token.endswith('>') or token.isupper():
         return [Argument(tokens.move())]
@@ -396,9 +403,7 @@ def parse_args(source, options):
     parsed = []
     while tokens.current() is not None:
         if tokens.current() == '--':
-            tokens.move()
-            parsed += [Argument(None, v) for v in tokens]
-            break
+            return parsed + [Argument(None, v) for v in tokens]
         elif tokens.current().startswith('--'):
             parsed += parse_long(tokens, options)
         elif tokens.current().startswith('-') and tokens.current() != '-':
@@ -413,8 +418,12 @@ def parse_doc_options(doc):
 
 
 def printable_usage(doc):
-    return re.split(r'\n\s*\n',
-            ''.join(re.split(r'([Uu][Ss][Aa][Gg][Ee]:)', doc)[1:]))[0].strip()
+    usage_split = re.split(r'([Uu][Ss][Aa][Gg][Ee]:)', doc)
+    if len(usage_split) < 3:
+        raise DocoptLanguageError('"usage:" (case-insensitive) not found.')
+    if len(usage_split) > 3:
+        raise DocoptLanguageError('More than one "usage:" (case-insensitive).')
+    return re.split(r'\n\s*\n', ''.join(usage_split[1:]))[0].strip()
 
 
 def formal_usage(printable_usage):
@@ -426,7 +435,7 @@ def extras(help, version, options, doc):
     if help and any((o.name in ('-h', '--help')) and o.value for o in options):
         print(doc.strip())
         exit()
-    if version and any(o.long == '--version' and o.value for o in options):
+    if version and any(o.name == '--version' and o.value for o in options):
         print(version)
         exit()
 
@@ -439,14 +448,14 @@ class Dict(dict):
 def docopt(doc, argv=sys.argv[1:], help=True, version=None):
     DocoptExit.usage = docopt.usage = usage = printable_usage(doc)
     pot_options = parse_doc_options(doc)
-    argv = parse_args(argv, options=pot_options)
-    options = [o for o in argv if type(o) is Option]
-    extras(help, version, options, doc)
     formal_pattern = parse_pattern(formal_usage(usage), options=pot_options)
-    pot_arguments = [a for a in formal_pattern.flat
-                     if type(a) in [Argument, Command]]
+    argv = parse_args(argv, options=pot_options)
+    extras(help, version, argv, doc)
     matched, left, arguments = formal_pattern.fix().match(argv)
     if matched and left == []:  # better message if left?
+        options = [o for o in argv if type(o) is Option]
+        pot_arguments = [a for a in formal_pattern.flat
+                         if type(a) in [Argument, Command]]
         return Dict((a.name, a.value) for a in
                     (pot_options + options + pot_arguments + arguments))
     raise DocoptExit()
